<h3> Get Larch</h3>
<p>Current version: <b>{{ release }}</b></p>

&nbsp; &nbsp; <a href="{{ pathto('installation') }}">Install Latest Version</a>  <br>
&nbsp; &nbsp; <a href="https://github.com/xraypy/xraylarch"> Development Version</a> <br>
&nbsp; &nbsp; <a href="{{ pathto('installation') }}#getting-help"> Getting Help</a><br>

<h3>Contents</h3>

&nbsp; &nbsp; <a href="{{ pathto('index') }}contents"> Table of Contents</a><br>

&nbsp; &nbsp; <a href="{{ pathto('getting_started') }}"> Getting Started</a><br>

&nbsp; &nbsp; <a href="{{ pathto('larix') }}"> Larix (was XAS Viewer) </a><br>
&nbsp; &nbsp; <a href="{{ pathto('guis') }}"> Other GUIs </a><br>
&nbsp; &nbsp; <a href="{{ pathto('data') }}"> Reading Data </a><br>
&nbsp; &nbsp; <a href="{{ pathto('plotting') }}"> Plotting Data</a><br>
&nbsp; &nbsp; <a href="{{ pathto('fitting') }}"> Fitting Data</a><br>
&nbsp; &nbsp; <a href="{{ pathto('xafs') }}"> XAFS Analysis </a><br>
&nbsp; &nbsp; <a href="{{ pathto('xray') }}"> X-ray Databases </a><br>
&nbsp; &nbsp; <a href="{{ pathto('xrf') }}"> X-ray Fluoresence Analysis</a><br>
&nbsp; &nbsp; <a href="{{ pathto('python') }}"> Larch Python library</a><br>
&nbsp; &nbsp; <a href="{{ pathto('larchlang') }}"> Larch macro language</a><br>
