@echo off

set Path="C:\Program Files\Microsoft Visual Studio 10.0\VC\bin";C:\WINDOWS\system32;C:\WINDOWS;C:\WINDOWS\System32\Wbem;C:\Python27;%PATH%


"C:\Program Files\Microsoft Visual Studio 10.0\VC\bin\vcvars32.bat"
"C:\Program Files\Microsoft Visual Studio 10.0\VC\bin\vcvarsall.bat"