# Sphinx extensions for generating PDF/EPUB output
extensions = ['sphinx.ext.imgmath',
              'sphinx.ext.autodoc',
              'sphinx.ext.napoleon',
              'sphinx.ext.extlinks',
              'sphinx.ext.intersphinx',
              'sphinx.ext.ifconfig',
              'sphinxcontrib.bibtex',
              'subfig']
