#!/usr/bin/env python

"""
Standard plots for Larch
------------------------

This module is independent of the Wx GUI. It uses matplotlib or Plotly as
plotting libraries. It is intended as plotter module for standalone scripts and
Jupyter notebooks.
"""

from . import plotly_xafsplots
